@@ -30,7 +30,7 @@
 #include <sys/wait.h> /* waitpid() */
 #endif
 #include <signal.h> /* SIG* kill() */
-#include <unistd.h> /* execve() fork() setsid() */
+#include <unistd.h> /* execve() fork() setsid() usleep() */
 
 #include <assert.h> /* assert() */
 #include <errno.h> /* errno */
@@ -40,6 +40,7 @@
 #include <string.h> /* strdup() */
 
 #include "cfg/config.h"
+#include "compat/os.h"
 #include "compat/pthread.h"
 #include "engine/var.h"
 #include "engine/variables.h"
@@ -49,6 +50,7 @@
 #include "ui/ui.h"
 #include "utils/cancellation.h"
 #include "utils/env.h"
+#include "utils/event.h"
 #include "utils/fs.h"
 #include "utils/log.h"
 #include "utils/path.h"
@@ -113,7 +115,7 @@ typedef struct
 background_task_args;
 
 static void set_jobcount_var(int count);
-static void job_check(bg_job_t *job);
+static void show_job_errors(bg_job_t *job);
 static void job_free(bg_job_t *job);
 static void * error_thread(void *p);
 static void update_error_jobs(bg_job_t **jobs);
@@ -125,22 +127,29 @@ static void rip_children(void);
 static void rip_child(pid_t pid, int status);
 static void report_error_msg(const char title[], const char text[]);
 #endif
-static bg_job_t * launch_external(const char cmd[], BgJobFlags flags,
-		ShellRequester by);
+static bg_job_t * launch_external(const char cmd[], const char pwd[],
+		BgJobFlags flags, ShellRequester by);
+#ifdef _WIN32
+static int finish_startup_info(STARTUPINFOW *startup);
+#endif
 static void append_error_msg(bg_job_t *job, const char err_msg[]);
 static void place_on_job_bar(bg_job_t *job);
 static void get_off_job_bar(bg_job_t *job);
 static bg_job_t * add_background_job(pid_t pid, const char cmd[],
 		uintptr_t err, uintptr_t data, BgJobType type, int with_bg_op);
 static void * background_task_bootstrap(void *arg);
-static void set_current_job(bg_job_t *job);
-static void make_current_job_key(void);
 static int update_job_status(bg_job_t *job);
 static void mark_job_finished(bg_job_t *job, int exit_code);
+static void maybe_wake_error_thread(void);
+static int is_job_erroring(bg_job_t *job);
+static void wake_error_thread(void);
 static int bg_op_cancel(bg_op_t *bg_op);
 
 bg_job_t *bg_jobs = NULL;
 
+/* Event to wake up error thread from sleep for processing by
+ * wake_error_thread(). */
+static event_t *error_thread_event;
 /* Head of list of newly started jobs. */
 static bg_job_t *new_err_jobs;
 /* Mutex to protect new_err_jobs. */
@@ -148,35 +157,61 @@ static pthread_mutex_t new_err_jobs_lock = PTHREAD_MUTEX_INITIALIZER;
 /* Conditional variable to signal availability of new jobs in new_err_jobs. */
 static pthread_cond_t new_err_jobs_cond = PTHREAD_COND_INITIALIZER;
 
-/* Thread local storage for bg_job_t associated with active thread. */
+/* Thread-local storage for bg_job_t associated with active thread. */
 static pthread_key_t current_job;
 
-void
+int
 bg_init(void)
 {
-	pthread_t id;
-	int err = pthread_create(&id, NULL, &error_thread, NULL);
-	assert(err == 0);
-	(void)err;
+	error_thread_event = event_alloc();
+	if(error_thread_event == NULL)
+	{
+		return 1;
+	}
 
-	/* Initialize state for the main thread. */
-	set_current_job(NULL);
+	/* Create thread-local storage before starting any background threads. */
+	if(pthread_key_create(&current_job, NULL) != 0)
+	{
+		event_free(error_thread_event);
+		return 1;
+	}
+
+	pthread_t id;
+	if(pthread_create(&id, NULL, &error_thread, NULL) != 0)
+	{
+		pthread_key_delete(current_job);
+		event_free(error_thread_event);
+		return 1;
+	}
+
+	return 0;
 }
 
 void
-bg_check(void)
+bg_check(int show_errors)
 {
+	static int checking;
+	if(checking)
+	{
+		/* This function is not re-entrant. */
+		return;
+	}
+
+	checking = 1;
+
 #ifndef _WIN32
+	/*
+	 * Rip children even if there are no jobs because it doesn't guarantee absence
+	 * of zombies.
+	 *
+	 * Do not do this in nested calls because implementation relies on job list
+	 * and won't be able to update job status if the list is not available leaving
+	 * job instances around in a permanent "running" state.
+	 */
 	rip_children();
 #endif
 
-	/* Quit if there is no jobs or list is unavailable (e.g. used by another
-	 * invocation of this function). */
-	if(bg_jobs == NULL)
-	{
-		set_jobcount_var(0);
-		return;
-	}
+	maybe_wake_error_thread();
 
 	int active_jobs = 0;
 
@@ -187,18 +222,38 @@ bg_check(void)
 	bg_job_t *prev = NULL;
 	while(p != NULL)
 	{
-		job_check(p);
-
-		pthread_spin_lock(&p->status_lock);
-		int running = p->running;
-		int can_remove = (!p->running && p->use_count == 0);
-		pthread_spin_unlock(&p->status_lock);
-
-		active_jobs += (running != 0 && p->in_menu);
-
-		if(!running && p->on_job_bar)
+		if(show_errors)
 		{
-			get_off_job_bar(p);
+			show_job_errors(p);
+		}
+
+		/* Checks status of the job (exit code isn't of much use here). */
+		(void)update_job_status(p);
+
+		/* In case of lock failure, assume the job is active. */
+		int running = 1;
+		int can_remove = 0;
+
+		if(pthread_spin_lock(&p->status_lock) == 0)
+		{
+			running = p->running;
+			can_remove = (!running && p->use_count == 0);
+			(void)pthread_spin_unlock(&p->status_lock);
+		}
+
+		active_jobs += (running && p->in_menu);
+
+		if(!running)
+		{
+			if(p->on_job_bar)
+			{
+				get_off_job_bar(p);
+			}
+			if(p->exit_cb != NULL)
+			{
+				p->exit_cb(p, p->exit_cb_arg);
+				p->exit_cb = NULL;
+			}
 		}
 
 		/* Remove job if it is finished now. */
@@ -225,6 +280,8 @@ bg_check(void)
 	bg_jobs = head;
 
 	set_jobcount_var(active_jobs);
+
+	checking = 0;
 }
 
 /* Updates builtin variable that holds number of active jobs.  Schedules UI
@@ -243,21 +300,23 @@ set_jobcount_var(int count)
 	}
 }
 
-/* Checks status of the job.  Processes error stream or checks whether process
- * is still running. */
+/* Processes error stream or checks whether process is still running. */
 static void
-job_check(bg_job_t *job)
+show_job_errors(bg_job_t *job)
 {
 	char *new_errors;
 
 	/* Display portions of errors from the job while there are any. */
 	do
 	{
-		pthread_spin_lock(&job->errors_lock);
-		new_errors = job->new_errors;
-		job->new_errors = NULL;
-		job->new_errors_len = 0U;
-		pthread_spin_unlock(&job->errors_lock);
+		new_errors = NULL;
+		if(pthread_spin_lock(&job->errors_lock) == 0)
+		{
+			new_errors = job->new_errors;
+			job->new_errors = NULL;
+			job->new_errors_len = 0U;
+			(void)pthread_spin_unlock(&job->errors_lock);
+		}
 
 		if(new_errors != NULL && !job->skip_errors)
 		{
@@ -267,8 +326,6 @@ job_check(bg_job_t *job)
 		free(new_errors);
 	}
 	while(new_errors != NULL);
-
-	(void)update_job_status(job);
 }
 
 /* Frees resources allocated by the job as well as the bg_job_t structure
@@ -377,7 +434,7 @@ bg_and_wait_for_errors(char cmd[], const struct cancellation_t *cancellation)
 		}
 		else
 		{
-			result = status_to_exit_code(get_proc_exit_status(pid));
+			result = status_to_exit_code(get_proc_exit_status(pid, cancellation));
 		}
 	}
 
@@ -405,14 +462,22 @@ error_thread(void *p)
 	(void)pthread_detach(pthread_self());
 	block_all_thread_signals();
 
+	const event_end_t event_end = event_wait_end(error_thread_event);
+
 	while(1)
 	{
 		update_error_jobs(&jobs);
 		make_ready_list(jobs, selector);
+		selector_add(selector, event_end);
 		while(selector_wait(selector, ERROR_SELECT_TIMEOUT_MS))
 		{
 			int need_update_list = (jobs == NULL);
 
+			if(selector_is_ready(selector, event_end))
+			{
+				(void)event_reset(error_thread_event);
+			}
+
 			bg_job_t **job = &jobs;
 			while(*job != NULL)
 			{
@@ -420,6 +485,15 @@ error_thread(void *p)
 				char err_msg[ERR_MSG_LEN];
 				ssize_t nread;
 
+				if(j->drained)
+				{
+					/* List update drops jobs which aren't running anymore thus allowing
+					 * them to be gone.  Matters at least in tests which wait for all
+					 * tasks to finish and looping here leads to a timeout. */
+					need_update_list = 1;
+					goto next_job;
+				}
+
 				if(!selector_is_ready(selector, j->err_stream))
 				{
 					goto next_job;
@@ -436,35 +510,26 @@ error_thread(void *p)
 					nread = bytes_read;
 				}
 #endif
-				if(nread < 0)
+				if(nread > 0)
 				{
+					err_msg[nread] = '\0';
+					append_error_msg(j, err_msg);
+				}
+				else
+				{
+					/* EOF or some error. */
 					need_update_list = 1;
 					j->drained = 1;
-					goto next_job;
 				}
 
-				if(nread == 0)
-				{
-					/* Reached EOF, exclude corresponding file descriptor from the set,
-					 * cut the job out of our list and decrement its use counter. */
-					selector_remove(selector, j->err_stream);
-					*job = j->err_next;
-					bg_job_decref(j);
-					continue;
-				}
-
-				err_msg[nread] = '\0';
-				append_error_msg(j, err_msg);
-
 			next_job:
 				job = &j->err_next;
 			}
 
-			if(!need_update_list)
+			if(!need_update_list && pthread_mutex_lock(&new_err_jobs_lock) == 0)
 			{
-				pthread_mutex_lock(&new_err_jobs_lock);
 				need_update_list = (new_err_jobs != NULL);
-				pthread_mutex_unlock(&new_err_jobs_lock);
+				(void)pthread_mutex_unlock(&new_err_jobs_lock);
 			}
 			if(need_update_list)
 			{
@@ -474,6 +539,7 @@ error_thread(void *p)
 	}
 
 	selector_free(selector);
+	event_free(error_thread_event);
 	return NULL;
 }
 
@@ -494,18 +560,15 @@ free_drained_jobs(bg_job_t **jobs)
 	{
 		bg_job_t *const j = *job;
 
-		if(j->drained)
+		if(j->drained && pthread_spin_lock(&j->status_lock) == 0)
 		{
-			pthread_spin_lock(&j->status_lock);
-			/* If finished, decrement use_count and drop it from the list. */
-			if(!j->running)
-			{
-				--j->use_count;
-				*job = j->err_next;
-				pthread_spin_unlock(&j->status_lock);
-				continue;
-			}
-			pthread_spin_unlock(&j->status_lock);
+			/* Drop it from the list even if the job is still running, we won't be
+			 * able to get anything out of it anyway. */
+			--j->use_count;
+			j->erroring = 0;
+			*job = j->err_next;
+			(void)pthread_spin_unlock(&j->status_lock);
+			continue;
 		}
 
 		job = &j->err_next;
@@ -519,14 +582,20 @@ import_error_jobs(bg_job_t **jobs)
 	bg_job_t *new_jobs;
 
 	/* Add new tasks to internal list, wait if there are no jobs. */
-	pthread_mutex_lock(&new_err_jobs_lock);
+	if(pthread_mutex_lock(&new_err_jobs_lock) != 0)
+	{
+		return;
+	}
 	while(*jobs == NULL && new_err_jobs == NULL)
 	{
-		pthread_cond_wait(&new_err_jobs_cond, &new_err_jobs_lock);
+		if(pthread_cond_wait(&new_err_jobs_cond, &new_err_jobs_lock) != 0)
+		{
+			break;
+		}
 	}
 	new_jobs = new_err_jobs;
 	new_err_jobs = NULL;
-	pthread_mutex_unlock(&new_err_jobs_lock);
+	(void)pthread_mutex_unlock(&new_err_jobs_lock);
 
 	/* Prepend new jobs to the list. */
 	while(new_jobs != NULL)
@@ -617,12 +686,11 @@ report_error_msg(const char title[], const char text[])
 static void
 append_error_msg(bg_job_t *job, const char err_msg[])
 {
-	if(err_msg[0] != '\0')
+	if(err_msg[0] != '\0' && pthread_spin_lock(&job->errors_lock) == 0)
 	{
-		pthread_spin_lock(&job->errors_lock);
 		(void)strappend(&job->errors, &job->errors_len, err_msg);
 		(void)strappend(&job->new_errors, &job->new_errors_len, err_msg);
-		pthread_spin_unlock(&job->errors_lock);
+		(void)pthread_spin_unlock(&job->errors_lock);
 	}
 }
 
@@ -634,17 +702,20 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 	int out_pipe[2];
 	int error_pipe[2];
 
-	if(pipe(out_pipe) != 0)
+	if(out != NULL && pipe(out_pipe) != 0)
 	{
 		show_error_msg("File pipe error", "Error creating pipe");
 		return (pid_t)-1;
 	}
 
-	if(pipe(error_pipe) != 0)
+	if(err != NULL && pipe(error_pipe) != 0)
 	{
 		show_error_msg("File pipe error", "Error creating pipe");
-		close(out_pipe[0]);
-		close(out_pipe[1]);
+		if(out != NULL)
+		{
+			close(out_pipe[0]);
+			close(out_pipe[1]);
+		}
 		return (pid_t)-1;
 	}
 
@@ -655,27 +726,29 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 
 	if((pid = fork()) == -1)
 	{
-		close(out_pipe[0]);
-		close(out_pipe[1]);
-		close(error_pipe[0]);
-		close(error_pipe[1]);
+		if(out != NULL)
+		{
+			close(out_pipe[0]);
+			close(out_pipe[1]);
+		}
+		if(err != NULL)
+		{
+			close(error_pipe[0]);
+			close(error_pipe[1]);
+		}
 		return (pid_t)-1;
 	}
 
 	if(pid == 0)
 	{
-		char *sh;
-		char *sh_flag;
-
-		close(out_pipe[0]);
-		close(error_pipe[0]);
-		if(dup2(out_pipe[1], STDOUT_FILENO) == -1)
+		if(out != NULL)
 		{
-			_Exit(EXIT_FAILURE);
+			bind_pipe_or_die(STDOUT_FILENO, out_pipe[1], out_pipe[0]);
 		}
-		if(dup2(error_pipe[1], STDERR_FILENO) == -1)
+
+		if(err != NULL)
 		{
-			_Exit(EXIT_FAILURE);
+			bind_pipe_or_die(STDERR_FILENO, error_pipe[1], error_pipe[0]);
 		}
 
 		if(in != NULL)
@@ -690,25 +763,34 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 			fclose(in);
 		}
 
-		sh = user_sh ? get_execv_path(cfg.shell) : "/bin/sh";
-		sh_flag = user_sh ? cfg.shell_cmd_flag : "-c";
+		char *sh_flag = user_sh ? cfg.shell_cmd_flag : "-c";
 		prepare_for_exec();
-		execvp(sh, make_execv_array(sh, sh_flag, cmd));
+		execvp(get_execv_path(cfg.shell),
+				make_execv_array(cfg.shell, sh_flag, cmd));
+		LOG_SERROR_MSG(errno, "Failed to launch a shell: `%s` `%s` `%s`", cfg.shell,
+				sh_flag, cmd);
 		_Exit(127);
 	}
 
-	close(out_pipe[1]);
-	close(error_pipe[1]);
-	*out = fdopen(out_pipe[0], "r");
-	*err = fdopen(error_pipe[0], "r");
+	if(out != NULL)
+	{
+		close(out_pipe[1]);
+		*out = fdopen(out_pipe[0], "r");
+	}
+
+	if(err != NULL)
+	{
+		close(error_pipe[1]);
+		*err = fdopen(error_pipe[0], "r");
+	}
 
 	return pid;
 }
 #else
 /* Runs command in a background and redirects its stdout and stderr streams to
- * file streams which are set.  Input is redirected only if in parameter isn't
- * NULL.  Don't pass pipe for input, it can cause deadlock.  Returns (pid_t)0 or
- * (pid_t)-1 on error. */
+ * file streams which are set.  Input and output are redirected only if the
+ * corresponding parameter isn't NULL.  Don't pass pipe for input, it can cause
+ * deadlock.  Returns (pid_t)0 or (pid_t)-1 on error. */
 static pid_t
 background_and_capture_internal(char cmd[], int user_sh, FILE *in, FILE **out,
 		FILE **err, int out_pipe[2], int err_pipe[2])
@@ -730,9 +812,9 @@ background_and_capture_internal(char cmd[], int user_sh, FILE *in, FILE **out,
 			return (pid_t)-1;
 	}
 
-	if(_dup2(out_pipe[1], _fileno(stdout)) != 0)
+	if(out != NULL && _dup2(out_pipe[1], _fileno(stdout)) != 0)
 		return (pid_t)-1;
-	if(_dup2(err_pipe[1], _fileno(stderr)) != 0)
+	if(err != NULL && _dup2(err_pipe[1], _fileno(stderr)) != 0)
 		return (pid_t)-1;
 
 	/* At least cmd.exe is incapable of handling UNC paths. */
@@ -781,11 +863,17 @@ background_and_capture_internal(char cmd[], int user_sh, FILE *in, FILE **out,
 		return (pid_t)-1;
 	}
 
-	if((*out = _fdopen(out_pipe[0], "r")) == NULL)
-		return (pid_t)-1;
-	if((*err = _fdopen(err_pipe[0], "r")) == NULL)
+	if(out != NULL && (*out = _fdopen(out_pipe[0], "r")) == NULL)
 	{
-		fclose(*out);
+		return (pid_t)-1;
+	}
+
+	if(err != NULL && (*err = _fdopen(err_pipe[0], "r")) == NULL)
+	{
+		if(out != NULL)
+		{
+			fclose(*out);
+		}
 		return (pid_t)-1;
 	}
 
@@ -800,17 +888,20 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 	int err_fd, err_pipe[2];
 	pid_t pid;
 
-	if(_pipe(out_pipe, 512, O_NOINHERIT) != 0)
+	if(out != NULL && _pipe(out_pipe, 512, O_NOINHERIT) != 0)
 	{
 		show_error_msg("File pipe error", "Error creating pipe");
 		return (pid_t)-1;
 	}
 
-	if(_pipe(err_pipe, 512, O_NOINHERIT) != 0)
+	if(err != NULL && _pipe(err_pipe, 512, O_NOINHERIT) != 0)
 	{
 		show_error_msg("File pipe error", "Error creating pipe");
-		close(out_pipe[0]);
-		close(out_pipe[1]);
+		if(out != NULL)
+		{
+			close(out_pipe[0]);
+			close(out_pipe[1]);
+		}
 		return (pid_t)-1;
 	}
 
@@ -821,8 +912,14 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 	pid = background_and_capture_internal(cmd, user_sh, in, out, err, out_pipe,
 			err_pipe);
 
-	_close(out_pipe[1]);
-	_close(err_pipe[1]);
+	if(out != NULL)
+	{
+		_close(out_pipe[1]);
+	}
+	if(err != NULL)
+	{
+		_close(err_pipe[1]);
+	}
 
 	_dup2(in_fd, _fileno(stdin));
 	_dup2(out_fd, _fileno(stdout));
@@ -830,8 +927,14 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 
 	if(pid == (pid_t)-1)
 	{
-		_close(out_pipe[0]);
-		_close(err_pipe[0]);
+		if(out != NULL)
+		{
+			_close(out_pipe[0]);
+		}
+		if(err != NULL)
+		{
+			_close(err_pipe[0]);
+		}
 	}
 
 	return pid;
@@ -839,8 +942,8 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 #endif
 
 int
-bg_run_external(const char cmd[], int skip_errors, ShellRequester by,
-		FILE **input)
+bg_run_external(const char cmd[], int keep_in_fg, int skip_errors,
+		ShellRequester by, FILE **input)
 {
 	char *command = (cfg.fast_run ? fast_run_complete(cmd) : strdup(cmd));
 	if(command == NULL)
@@ -848,8 +951,10 @@ bg_run_external(const char cmd[], int skip_errors, ShellRequester by,
 		return 1;
 	}
 
-	const BgJobFlags flags = (input == NULL ? BJF_NONE : BJF_SUPPLY_INPUT);
-	bg_job_t *job = launch_external(command, flags, by);
+	BgJobFlags flags = (keep_in_fg ? BJF_KEEP_IN_FG : BJF_NONE)
+	                 | (input != NULL ? BJF_SUPPLY_INPUT : BJF_NONE);
+
+	bg_job_t *job = launch_external(command, /*pwd=*/NULL, flags, by);
 	free(command);
 	if(job == NULL)
 	{
@@ -873,9 +978,10 @@ bg_run_external(const char cmd[], int skip_errors, ShellRequester by,
 }
 
 bg_job_t *
-bg_run_external_job(const char cmd[], BgJobFlags flags)
+bg_run_external_job(const char cmd[], BgJobFlags flags, const char descr[],
+		const char pwd[])
 {
-	bg_job_t *job = launch_external(cmd, flags, SHELL_BY_APP);
+	bg_job_t *job = launch_external(cmd, pwd, flags, SHELL_BY_APP);
 	if(job == NULL)
 	{
 		return NULL;
@@ -888,6 +994,12 @@ bg_run_external_job(const char cmd[], BgJobFlags flags)
 
 	if(flags & BJF_JOB_BAR_VISIBLE)
 	{
+		/* Set description before placing the job on the bar so that the first
+		 * redraw will already have the description. */
+		if(descr != NULL)
+		{
+			bg_op_set_descr(&job->bg_op, descr);
+		}
 		place_on_job_bar(job);
 	}
 
@@ -896,9 +1008,11 @@ bg_run_external_job(const char cmd[], BgJobFlags flags)
 	return job;
 }
 
-/* Starts a new external command job.  Returns the new job or NULL on error. */
+/* Starts a new external command job.  pwd can be NULL, otherwise it should be
+ * a valid path.  Returns the new job or NULL on error. */
 static bg_job_t *
-launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
+launch_external(const char cmd[], const char pwd[], BgJobFlags flags,
+		ShellRequester by)
 {
 	/* TODO: simplify this function (launch_external()) somehow, maybe split in
 	 *       two. */
@@ -908,9 +1022,11 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 	const int merge_streams = (capture_output && (flags & BJF_MERGE_STREAMS));
 
 #ifndef _WIN32
+	const int keep_in_fg = (flags & BJF_KEEP_IN_FG);
+
 	pid_t pid;
-	int input_pipe[2];
-	int output_pipe[2];
+	int input_pipe[2] = { -1, -1 };
+	int output_pipe[2] = { -1, -1 };
 
 	/* For the sake of simplicity just use -1, calling close(-1) won't hurt. */
 	int error_pipe[2] = { -1, -1 };
@@ -920,6 +1036,13 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 		return NULL;
 	}
 
+	if(pwd != NULL && (!is_dir(pwd) || os_access(pwd, X_OK) != 0))
+	{
+		/* CreateProcessW() on Windows fails in this case resulting in the function
+		 * returning NULL, do the same here for consistent behaviour. */
+		return NULL;
+	}
+
 	if(supply_input)
 	{
 		if(pipe(input_pipe) != 0)
@@ -965,6 +1088,12 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 	{
 		extern char **environ;
 
+		if(pwd != NULL && chdir(pwd) != 0)
+		{
+			perror("chdir");
+			_Exit(EXIT_FAILURE);
+		}
+
 		int stderr_pipe = (merge_streams ? output_pipe[1] : error_pipe[1]);
 
 		/* Redirect stderr to write end of pipe. */
@@ -975,32 +1104,22 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 		}
 		close(STDIN_FILENO);
 		close(STDOUT_FILENO);
-		/* Close read end of pipe. */
+
+		/* Close original error pipe descriptors. */
 		if(error_pipe[0] != -1)
 		{
 			close(error_pipe[0]);
+			close(error_pipe[1]);
 		}
 
 		if(supply_input)
 		{
-			if(dup2(input_pipe[0], STDIN_FILENO) == -1)
-			{
-				perror("dup2");
-				_Exit(EXIT_FAILURE);
-			}
-			/* Close write end of pipe. */
-			close(input_pipe[1]);
+			bind_pipe_or_die(STDIN_FILENO, input_pipe[0], input_pipe[1]);
 		}
 
 		if(capture_output)
 		{
-			if(dup2(output_pipe[1], STDOUT_FILENO) == -1)
-			{
-				perror("dup2");
-				_Exit(EXIT_FAILURE);
-			}
-			/* Close read end of pipe. */
-			close(output_pipe[0]);
+			bind_pipe_or_die(STDOUT_FILENO, output_pipe[1], output_pipe[0]);
 		}
 
 		/* Attach stdin and optionally stdout to /dev/null. */
@@ -1017,11 +1136,15 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 				perror("dup2 for stdout");
 				_Exit(EXIT_FAILURE);
 			}
+			if(nullfd != STDIN_FILENO && nullfd != STDOUT_FILENO)
+			{
+				close(nullfd);
+			}
 		}
 
 		/* setsid() creates process group as well and doesn't work if current
 		 * process is a group leader, so don't do setpgid(). */
-		if(setsid() == (pid_t)-1)
+		if(!keep_in_fg && setsid() == (pid_t)-1)
 		{
 			perror("setsid");
 			_Exit(EXIT_FAILURE);
@@ -1031,6 +1154,8 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 		char *sh_flag = (by == SHELL_BY_USER ? cfg.shell_cmd_flag : "-c");
 		execve(get_execv_path(cfg.shell),
 				make_execv_array(cfg.shell, sh_flag, strdup(cmd)), environ);
+		LOG_SERROR_MSG(errno, "Failed to launch a shell: `%s` `%s` `%s`", cfg.shell,
+				sh_flag, cmd);
 		_Exit(127);
 	}
 
@@ -1071,32 +1196,30 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 
 	return job;
 #else
-	STARTUPINFOW startup = { .dwFlags = STARTF_USESTDHANDLES };
+	/* Handles are either set below or redirected to NUL in
+	 * finish_startup_info(). */
+	STARTUPINFOW startup = {
+		.dwFlags = STARTF_USESTDHANDLES,
+		.hStdInput = INVALID_HANDLE_VALUE,
+		.hStdOutput = INVALID_HANDLE_VALUE,
+		.hStdError = INVALID_HANDLE_VALUE,
+	};
 	PROCESS_INFORMATION pinfo;
 	char *sh_cmd;
-	wchar_t *wide_cmd;
-
-	HANDLE hnul = CreateFileA("NUL", GENERIC_READ | GENERIC_WRITE, 0, NULL,
-			OPEN_EXISTING, 0, NULL);
-	if(hnul == INVALID_HANDLE_VALUE)
-	{
-		return NULL;
-	}
-	startup.hStdInput = hnul;
-	startup.hStdOutput = hnul;
 
 	HANDLE herr = INVALID_HANDLE_VALUE;
 	if(!merge_streams && !CreatePipe(&herr, &startup.hStdError, NULL, 16*1024))
 	{
-		CloseHandle(hnul);
 		return NULL;
 	}
 
 	HANDLE hin = INVALID_HANDLE_VALUE;
 	if(supply_input && !CreatePipe(&startup.hStdInput, &hin, NULL, 16*1024))
 	{
-		CloseHandle(herr);
-		CloseHandle(hnul);
+		if(herr != INVALID_HANDLE_VALUE)
+		{
+			CloseHandle(herr);
+		}
 		return NULL;
 	}
 
@@ -1105,42 +1228,58 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 	{
 		if(!CreatePipe(&hout, &startup.hStdOutput, NULL, 16*1024))
 		{
-			CloseHandle(herr);
+			if(herr != INVALID_HANDLE_VALUE)
+			{
+				CloseHandle(herr);
+			}
 			CloseHandle(hin);
-			CloseHandle(hnul);
 			return NULL;
 		}
 
 		if(merge_streams)
 		{
-			startup.hStdError = startup.hStdOutput;
+			/* Duplicate instead of just assigning so that closing one handle keeps
+			 * the other one operational. */
+			HANDLE this_process = GetCurrentProcess();
+			if(!DuplicateHandle(this_process, startup.hStdOutput, this_process,
+						&startup.hStdError, /*access=*/0, /*inheritable=*/TRUE,
+						DUPLICATE_SAME_ACCESS))
+			{
+				CloseHandle(startup.hStdOutput);
+				CloseHandle(hout);
+				CloseHandle(hin);
+				return NULL;
+			}
 		}
 	}
 
-	SetHandleInformation(startup.hStdInput, HANDLE_FLAG_INHERIT, 1);
-	SetHandleInformation(startup.hStdOutput, HANDLE_FLAG_INHERIT, 1);
-	SetHandleInformation(startup.hStdError, HANDLE_FLAG_INHERIT, 1);
-
+	finish_startup_info(&startup);
 	sh_cmd = win_make_sh_cmd(cmd, by);
 
-	wide_cmd = to_wide(sh_cmd);
-	int started = CreateProcessW(NULL, wide_cmd, NULL, NULL, 1, CREATE_SUSPENDED,
-			NULL, NULL, &startup, &pinfo);
+	wchar_t *wide_cmd = to_wide(sh_cmd);
+	wchar_t *wide_pwd = (pwd != NULL ? to_wide(pwd) : NULL);
+
+	int started = 0;
+	if(wide_cmd != NULL && (pwd == NULL || wide_pwd != NULL))
+	{
+		started = CreateProcessW(NULL, wide_cmd, NULL, NULL, 1, CREATE_SUSPENDED,
+				NULL, wide_pwd, &startup, &pinfo);
+	}
+
 	free(wide_cmd);
-	CloseHandle(hnul);
-	if(startup.hStdInput != INVALID_HANDLE_VALUE)
-	{
-		CloseHandle(startup.hStdInput);
-	}
+	free(wide_pwd);
+
+	CloseHandle(startup.hStdInput);
 	CloseHandle(startup.hStdOutput);
-	if(startup.hStdError != startup.hStdOutput)
-	{
-		CloseHandle(startup.hStdError);
-	}
+	CloseHandle(startup.hStdError);
 
 	if(!started)
 	{
 		free(sh_cmd);
+		if(herr != INVALID_HANDLE_VALUE)
+		{
+			CloseHandle(herr);
+		}
 		CloseHandle(hout);
 		CloseHandle(hin);
 		return NULL;
@@ -1158,7 +1297,10 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 
 	if(job == NULL)
 	{
-		CloseHandle(herr);
+		if(herr != INVALID_HANDLE_VALUE)
+		{
+			CloseHandle(herr);
+		}
 		CloseHandle(hin);
 		CloseHandle(hout);
 		CloseHandle(pinfo.hProcess);
@@ -1204,6 +1346,69 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 #endif
 }
 
+#ifdef _WIN32
+/* Makes sure that standard handles of startup structure which weren't
+ * initialized are redirected to NUL.  Returns zero on success. */
+static int
+finish_startup_info(STARTUPINFOW *startup)
+{
+	int missing = (startup->hStdInput == INVALID_HANDLE_VALUE)
+	            + (startup->hStdOutput == INVALID_HANDLE_VALUE)
+	            + (startup->hStdError == INVALID_HANDLE_VALUE);
+	if(missing == 0)
+	{
+		goto no_redirects;
+	}
+
+	/* Open up to three handles so that child process could close one of them
+	 * while keep using others. */
+	HANDLE hnul[3];
+	hnul[0] = CreateFileA("NUL", GENERIC_READ | GENERIC_WRITE,
+			FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
+			FILE_ATTRIBUTE_NORMAL, NULL);
+	if(hnul[0] == INVALID_HANDLE_VALUE)
+	{
+		return 1;
+	}
+
+	int i;
+	HANDLE this_process = GetCurrentProcess();
+	for(i = 1; i < missing; ++i)
+	{
+		if(!DuplicateHandle(this_process, hnul[0], this_process, &hnul[i],
+					/*access=*/0, /*inheritable=*/TRUE, DUPLICATE_SAME_ACCESS))
+		{
+			int j;
+			for(j = 0; j < i; ++j)
+			{
+				CloseHandle(hnul[j]);
+			}
+			return 1;
+		}
+	}
+
+	if(startup->hStdInput == INVALID_HANDLE_VALUE)
+	{
+		startup->hStdInput = hnul[--missing];
+	}
+	if(startup->hStdOutput == INVALID_HANDLE_VALUE)
+	{
+		startup->hStdOutput = hnul[--missing];
+	}
+	if(startup->hStdError == INVALID_HANDLE_VALUE)
+	{
+		startup->hStdError = hnul[--missing];
+	}
+
+no_redirects:
+	SetHandleInformation(startup->hStdInput, HANDLE_FLAG_INHERIT, 1);
+	SetHandleInformation(startup->hStdOutput, HANDLE_FLAG_INHERIT, 1);
+	SetHandleInformation(startup->hStdError, HANDLE_FLAG_INHERIT, 1);
+
+	return 0;
+}
+#endif
+
 int
 bg_execute(const char descr[], const char op_descr[], int total, int important,
 		bg_task_func task_func, void *args)
@@ -1240,10 +1445,12 @@ bg_execute(const char descr[], const char op_descr[], int total, int important,
 	if(pthread_create(&id, NULL, &background_task_bootstrap, task_args) != 0)
 	{
 		/* Mark job as finished with error. */
-		pthread_spin_lock(&task_args->job->status_lock);
-		task_args->job->running = 0;
-		task_args->job->exit_code = 1;
-		pthread_spin_unlock(&task_args->job->status_lock);
+		if(pthread_spin_lock(&task_args->job->status_lock) == 0)
+		{
+			task_args->job->running = 0;
+			task_args->job->exit_code = 1;
+			(void)pthread_spin_unlock(&task_args->job->status_lock);
+		}
 
 		free(task_args);
 		ret = 1;
@@ -1281,12 +1488,15 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 	bg_job_t *new = malloc(sizeof(*new));
 	if(new == NULL)
 	{
-		show_error_msg("Memory error", "Unable to allocate enough memory");
 		return NULL;
 	}
+	new->cmd = strdup(cmd);
+	if(new->cmd == NULL)
+	{
+		goto free_new;
+	}
 	new->type = type;
 	new->pid = pid;
-	new->cmd = strdup(cmd);
 	new->next = bg_jobs;
 	new->skip_errors = 0;
 	new->new_errors = NULL;
@@ -1295,15 +1505,31 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 	new->errors_len = 0U;
 	new->cancelled = 0;
 
-	pthread_spin_init(&new->errors_lock, PTHREAD_PROCESS_PRIVATE);
-	pthread_spin_init(&new->status_lock, PTHREAD_PROCESS_PRIVATE);
+	if(pthread_spin_init(&new->errors_lock, PTHREAD_PROCESS_PRIVATE) != 0)
+	{
+		goto free_cmd;
+	}
+	if(pthread_spin_init(&new->status_lock, PTHREAD_PROCESS_PRIVATE) != 0)
+	{
+		goto free_errors_lock;
+	}
+	if(with_bg_op &&
+			pthread_spin_init(&new->bg_op_lock, PTHREAD_PROCESS_PRIVATE) != 0)
+	{
+		goto free_status_lock;
+	}
+
 	new->running = 1;
+	new->erroring = 0;
 	new->use_count = 0;
 	new->exit_code = -1;
 
 	new->input = NULL;
 	new->output = NULL;
 
+	new->exit_cb = NULL;
+	new->exit_cb_arg = NULL;
+
 #ifndef _WIN32
 	new->err_stream = (int)err;
 #else
@@ -1314,20 +1540,21 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 
 	if(new->err_stream != NO_JOB_ID)
 	{
+		new->erroring = 1;
 		++new->use_count;
-		pthread_mutex_lock(&new_err_jobs_lock);
+
+		if(pthread_mutex_lock(&new_err_jobs_lock) != 0)
+		{
+			goto free_bg_op_lock;
+		}
 		new->err_next = new_err_jobs;
 		new_err_jobs = new;
-		pthread_mutex_unlock(&new_err_jobs_lock);
-		pthread_cond_signal(&new_err_jobs_cond);
+		(void)pthread_mutex_unlock(&new_err_jobs_lock);
+		(void)pthread_cond_signal(&new_err_jobs_cond);
 	}
 
 	new->with_bg_op = with_bg_op;
 	new->on_job_bar = 0;
-	if(new->with_bg_op)
-	{
-		pthread_spin_init(&new->bg_op_lock, PTHREAD_PROCESS_PRIVATE);
-	}
 	new->bg_op.total = 0;
 	new->bg_op.done = 0;
 	new->bg_op.progress = -1;
@@ -1338,6 +1565,21 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 
 	bg_jobs = new;
 	return new;
+
+free_bg_op_lock:
+	if(with_bg_op)
+	{
+		(void)pthread_spin_destroy(&new->bg_op_lock);
+	}
+free_status_lock:
+	(void)pthread_spin_destroy(&new->status_lock);
+free_errors_lock:
+	(void)pthread_spin_destroy(&new->errors_lock);
+free_cmd:
+	free(new->cmd);
+free_new:
+	free(new);
+	return NULL;
 }
 
 /* pthreads entry point for a new background task.  Performs correct
@@ -1350,35 +1592,22 @@ background_task_bootstrap(void *arg)
 
 	(void)pthread_detach(pthread_self());
 	block_all_thread_signals();
-	set_current_job(task_args->job);
 
-	task_args->func(&task_args->job->bg_op, task_args->args);
-
-	/* Mark task as finished normally. */
-	mark_job_finished(task_args->job, 0);
+	if(pthread_setspecific(current_job, task_args->job) == 0)
+	{
+		task_args->func(&task_args->job->bg_op, task_args->args);
+		mark_job_finished(task_args->job, /*exit_code=*/0);
+	}
+	else
+	{
+		mark_job_finished(task_args->job, /*exit_code=*/1);
+	}
 
 	free(task_args);
 
 	return NULL;
 }
 
-/* Stores pointer to the job in a thread-local storage. */
-static void
-set_current_job(bg_job_t *job)
-{
-	static pthread_once_t once = PTHREAD_ONCE_INIT;
-	pthread_once(&once, &make_current_job_key);
-
-	(void)pthread_setspecific(current_job, job);
-}
-
-/* current_job initializer for pthread_once(). */
-static void
-make_current_job_key(void)
-{
-	(void)pthread_key_create(&current_job, NULL);
-}
-
 int
 bg_has_active_jobs(int important_only)
 {
@@ -1396,6 +1625,13 @@ bg_has_active_jobs(int important_only)
 	return running;
 }
 
+void
+bg_job_set_exit_cb(bg_job_t *job, bg_job_exit_func cb, void *arg)
+{
+	job->exit_cb = cb;
+	job->exit_cb_arg = arg;
+}
+
 int
 bg_job_cancel(bg_job_t *job)
 {
@@ -1468,19 +1704,25 @@ bg_job_terminate(bg_job_t *job)
 int
 bg_job_is_running(bg_job_t *job)
 {
-	pthread_spin_lock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) != 0)
+	{
+		return 1;
+	}
 	int running = job->running;
-	pthread_spin_unlock(&job->status_lock);
+	(void)pthread_spin_unlock(&job->status_lock);
 	return (running && update_job_status(job));
 }
 
 int
 bg_job_was_killed(bg_job_t *job)
 {
-	pthread_spin_lock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) != 0)
+	{
+		return 0;
+	}
 	int running = job->running;
 	int exit_code = job->exit_code;
-	pthread_spin_unlock(&job->status_lock);
+	(void)pthread_spin_unlock(&job->status_lock);
 	return (!running && exit_code >= 0);
 }
 
@@ -1510,7 +1752,7 @@ bg_job_wait(bg_job_t *job)
 	}
 
 #ifndef _WIN32
-	int status = get_proc_exit_status(job->pid);
+	int status = get_proc_exit_status(job->pid, &no_cancellation);
 	if(status == -1)
 	{
 		return 1;
@@ -1557,34 +1799,117 @@ update_job_status(bg_job_t *job)
 static void
 mark_job_finished(bg_job_t *job, int exit_code)
 {
-	pthread_spin_lock(&job->status_lock);
-	job->running = 0;
-	job->exit_code = exit_code;
-	pthread_spin_unlock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) == 0)
+	{
+		job->running = 0;
+		job->exit_code = exit_code;
+		(void)pthread_spin_unlock(&job->status_lock);
+	}
+}
+
+int
+bg_job_wait_errors(bg_job_t *job)
+{
+	enum
+	{
+		ERROR_SLEEP_US = 50,
+		ERROR_SLEEP_MAX_US = 50*1000, /* 50ms should be more than enough. */
+	};
+
+	if(job->err_stream == NO_JOB_ID || bg_job_is_running(job))
+	{
+		return 0;
+	}
+
+	/* Using active polling with a sleep to avoid adding a mutex and a conditional
+	 * variable to every job with an error stream.  The code below shouldn't run
+	 * often. */
+
+	int i;
+	int erroring = 1;
+	for(i = 0; i < ERROR_SLEEP_MAX_US/ERROR_SLEEP_US && erroring; ++i)
+	{
+		erroring = is_job_erroring(job);
+		if(erroring)
+		{
+			wake_error_thread();
+			usleep(ERROR_SLEEP_US);
+		}
+	}
+
+	/* In case we've reached here and `erroring` is still non-zero, this could be
+	 * a bug in handling jobs or the system is under heavy load.  Either way, we
+	 * probably shouldn't wait here forever, so return an error. */
+	return erroring;
+}
+
+/* Wakes up error thread to process any changes to the jobs if it makes
+ * sense. */
+static void
+maybe_wake_error_thread(void)
+{
+	/* Don't wake up the error thread unless there is at least one job handled by
+	 * it. */
+	bg_job_t *job;
+	for(job = bg_jobs; job != NULL; job = job->next)
+	{
+		if(is_job_erroring(job))
+		{
+			wake_error_thread();
+			break;
+		}
+	}
+}
+
+/* Checks whether the job is being used by the error thread.  Returns non-zero
+ * if so. */
+static int
+is_job_erroring(bg_job_t *job)
+{
+	int erroring = 0;
+	if(pthread_spin_lock(&job->status_lock) == 0)
+	{
+		erroring = job->erroring;
+		(void)pthread_spin_unlock(&job->status_lock);
+	}
+	return erroring;
+}
+
+/* Wakes up error thread to process any changes to the jobs. */
+static void
+wake_error_thread(void)
+{
+	(void)event_signal(error_thread_event);
 }
 
 void
 bg_job_incref(bg_job_t *job)
 {
-	pthread_spin_lock(&job->status_lock);
-	++job->use_count;
-	pthread_spin_unlock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) == 0)
+	{
+		++job->use_count;
+		(void)pthread_spin_unlock(&job->status_lock);
+	}
 }
 
 void
 bg_job_decref(bg_job_t *job)
 {
-	pthread_spin_lock(&job->status_lock);
-	--job->use_count;
-	pthread_spin_unlock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) == 0)
+	{
+		--job->use_count;
+		assert(job->use_count >= 0 && "Excessive bg_job_decref() call!");
+		(void)pthread_spin_unlock(&job->status_lock);
+	}
 }
 
-void
+int
 bg_op_lock(bg_op_t *bg_op)
 {
 	bg_job_t *const job = STRUCT_FROM_FIELD(bg_job_t, bg_op, bg_op);
 	assert(job->with_bg_op && "This function requires bg_op data.");
-	pthread_spin_lock(&job->bg_op_lock);
+
+	return (pthread_spin_lock(&job->bg_op_lock) == 0);
 }
 
 void
@@ -1592,7 +1917,10 @@ bg_op_unlock(bg_op_t *bg_op)
 {
 	bg_job_t *const job = STRUCT_FROM_FIELD(bg_job_t, bg_op, bg_op);
 	assert(job->with_bg_op && "This function requires bg_op data.");
-	pthread_spin_unlock(&job->bg_op_lock);
+
+	int error = pthread_spin_unlock(&job->bg_op_lock);
+	assert(error == 0 && "Unlock failure in bg_op_lock()");
+	(void)error;
 }
 
 void
@@ -1604,11 +1932,13 @@ bg_op_changed(bg_op_t *bg_op)
 void
 bg_op_set_descr(bg_op_t *bg_op, const char descr[])
 {
-	bg_op_lock(bg_op);
-	replace_string(&bg_op->descr, descr);
-	bg_op_unlock(bg_op);
+	if(bg_op_lock(bg_op))
+	{
+		replace_string(&bg_op->descr, descr);
+		bg_op_unlock(bg_op);
 
-	bg_op_changed(bg_op);
+		bg_op_changed(bg_op);
+	}
 }
 
 /* Convenience method to cancel background job.  Returns previous version of the
@@ -1616,26 +1946,27 @@ bg_op_set_descr(bg_op_t *bg_op, const char descr[])
 static int
 bg_op_cancel(bg_op_t *bg_op)
 {
-	int was_cancelled;
+	int was_cancelled = 0;
+	if(bg_op_lock(bg_op))
+	{
+		was_cancelled = bg_op->cancelled;
+		bg_op->cancelled = 1;
+		bg_op_unlock(bg_op);
 
-	bg_op_lock(bg_op);
-	was_cancelled = bg_op->cancelled;
-	bg_op->cancelled = 1;
-	bg_op_unlock(bg_op);
-
-	bg_op_changed(bg_op);
+		bg_op_changed(bg_op);
+	}
 	return was_cancelled;
 }
 
 int
 bg_op_cancelled(bg_op_t *bg_op)
 {
-	int cancelled;
-
-	bg_op_lock(bg_op);
-	cancelled = bg_op->cancelled;
-	bg_op_unlock(bg_op);
-
+	int cancelled = 0;
+	if(bg_op_lock(bg_op))
+	{
+		cancelled = bg_op->cancelled;
+		bg_op_unlock(bg_op);
+	}
 	return cancelled;
 }
 
