@@ -27,6 +27,7 @@
 #include <fcntl.h>    /* O_RDWR, O_EXCL, O_CREAT, ... */
 #include <unistd.h>   /* ftruncate */
 
+#include "compat/fs_limits.h"
 #include "compat/reallocarray.h"
 #include "modes/dialogs/msg_dialog.h" /* show_error_msgf */
 #include "utils/fs.h"
@@ -338,7 +339,6 @@ regs_list(const char registers[])
 	{
 		reg_t *reg = reg_from_name(*registers++);
 		char reg_str[16];
-		int i;
 
 		if(reg == NULL || reg->nfiles <= 0)
 		{
@@ -348,8 +348,8 @@ regs_list(const char registers[])
 		snprintf(reg_str, sizeof(reg_str), "\"%c", reg->name);
 		len = add_to_string_array(&list, len, reg_str);
 
-		i = reg->nfiles;
-		while(i-- > 0)
+		int i;
+		for(i = 0; i < reg->nfiles; ++i)
 		{
 			len = add_to_string_array(&list, len, reg->files[i]);
 		}
