@@ -10,8 +10,8 @@
 
 #include "test.h"
 
-static void column_line_print(const void *data, int column_id, const char buf[],
-		size_t offset, AlignType align);
+static void column_line_print(const char buf[], int offset, AlignType align,
+		const format_info_t *info);
 static void column1_func(void *data, size_t buf_len, char buf[],
 		const format_info_t *info);
 static void column2_func(void *data, size_t buf_len, char buf[],
@@ -22,6 +22,7 @@ static const size_t MAX_WIDTH = 20;
 static char print_buffer[80 + 1];
 
 static const char *col1_str;
+static const char *col2_str;
 
 SETUP_ONCE()
 {
@@ -39,6 +40,7 @@ SETUP()
 	col2_next = &column2_func;
 
 	col1_str = "师从螺丝刀йклмнопрстуфхцчшщьыъэюя";
+	col2_str = "яюэъыьщшчцхфутсрпонмлкйизжёедгв推";
 }
 
 TEARDOWN()
@@ -49,8 +51,8 @@ TEARDOWN()
 }
 
 static void
-column_line_print(const void *data, int column_id, const char buf[],
-		size_t offset, AlignType align)
+column_line_print(const char buf[], int offset, AlignType align,
+		const format_info_t *info)
 {
 	memcpy(print_buffer + utf8_nstrsnlen(print_buffer, offset), buf, strlen(buf));
 }
@@ -64,7 +66,7 @@ column1_func(void *data, size_t buf_len, char buf[], const format_info_t *info)
 static void
 column2_func(void *data, size_t buf_len, char buf[], const format_info_t *info)
 {
-	snprintf(buf, buf_len + 1, "%s", "яюэъыьщшчцхфутсрпонмлкйизжёедгв推");
+	snprintf(buf, buf_len + 1, "%s", col2_str);
 }
 
 static void
@@ -140,6 +142,31 @@ TEST(none_cropping_allows_for_correct_gaps, IF(locale_works))
 	assert_string_equal(expected, print_buffer);
 }
 
+TEST(none_cropping_cuts_wide_strings_correctly, IF(locale_works))
+{
+	static column_info_t column_infos[2] = {
+		{ .column_id = COL1_ID, .full_width = 10UL,    .text_width = 10UL,
+		  .align = AT_LEFT,     .sizing = ST_AUTO,     .cropping = CT_NONE, },
+		{ .column_id = COL2_ID, .full_width = 4UL,     .text_width = 4UL,
+		  .align = AT_RIGHT,    .sizing = ST_AUTO,     .cropping = CT_NONE, },
+	};
+
+	col1_str = "师从螺丝";
+	col2_str = "xyz";
+
+	perform_test(column_infos, 2, 9);
+	assert_string_equal("师从螺xyz", print_buffer);
+
+	perform_test(column_infos, 2, 10);
+	assert_string_equal("师从螺 xyz", print_buffer);
+
+	perform_test(column_infos, 2, 11);
+	assert_string_equal("师从螺丝xyz", print_buffer);
+
+	perform_test(column_infos, 2, 12);
+	assert_string_equal("师从螺丝 xyz", print_buffer);
+}
+
 TEST(add_ellipsis_ok, IF(locale_works))
 {
 	static column_info_t column_infos[2] = {
@@ -227,6 +254,24 @@ TEST(wide_right_ellipsis_ok, IF(locale_works))
 	assert_string_equal(expected, print_buffer);
 }
 
+TEST(wide_middle_ellipsis_ok, IF(locale_works))
+{
+	/* A gap might appear after removing several characters to insert ellipsis in
+	 * their place, make sure that it's filled with spaces to obtain requested
+	 * width. */
+
+	static column_info_t column_infos[1] = {
+		{ .column_id = COL1_ID, .full_width = 0UL, .text_width = 0UL,
+		  .align = AT_MIDDLE,   .sizing = ST_AUTO, .cropping = CT_ELLIPSIS, },
+	};
+	static const char expected[] = ",,师从螺...从螺丝刀 ";
+
+	col1_str = ",,师从螺丝刀师从螺丝刀";
+	perform_test(column_infos, ARRAY_LEN(column_infos), MAX_WIDTH);
+
+	assert_string_equal(expected, print_buffer);
+}
+
 static int
 locale_works(void)
 {
