@@ -32,6 +32,7 @@
 #include "../../src/utils/macros.h"
 #include "../../src/utils/matcher.h"
 #include "../../src/utils/path.h"
+#include "../../src/utils/regexp.h"
 #include "../../src/utils/str.h"
 #include "../../src/utils/string_array.h"
 #include "../../src/utils/utils.h"
@@ -64,7 +65,7 @@ static void init_list(view_t *view);
 static int init_pair_stub(int pair, int f, int b);
 static int pair_content_stub(int pair, int *f, int *b);
 static int pair_in_use_stub(int pair);
-static void move_pair_stub(int from, int to);
+static void pair_moved_stub(int from, int to);
 
 void
 fix_environ(void)
@@ -98,6 +99,7 @@ conf_setup(void)
 	update_string(&cfg.media_prg, "");
 	update_string(&cfg.vborder_filler, "");
 	update_string(&cfg.hborder_filler, "");
+	update_string(&cfg.millersep_filler, "");
 	update_string(&cfg.tab_prefix, "");
 	update_string(&cfg.tab_label, "");
 	update_string(&cfg.tab_suffix, "");
@@ -108,6 +110,8 @@ conf_setup(void)
 	cfg.sizefmt.ieci_prefixes = 0;
 	cfg.sizefmt.space = 0;
 
+	cfg.tab_stop = 4;
+
 #ifndef _WIN32
 	replace_string(&cfg.shell, "/bin/sh");
 	update_string(&cfg.shell_cmd_flag, "-c");
@@ -118,6 +122,15 @@ conf_setup(void)
 	stats_update_shell_type(cfg.shell);
 
 	cfg.dot_dirs = DD_TREE_LEAFS_PARENT;
+
+	if(cfg.wild_inc != NULL)
+	{
+		matcher_free(cfg.wild_inc);
+	}
+
+	char *error;
+	cfg.wild_inc = matcher_alloc_glob("", &error);
+	free(error);
 }
 
 void
@@ -139,6 +152,7 @@ conf_teardown(void)
 	update_string(&cfg.media_prg, NULL);
 	update_string(&cfg.vborder_filler, NULL);
 	update_string(&cfg.hborder_filler, NULL);
+	update_string(&cfg.millersep_filler, NULL);
 	update_string(&cfg.tab_prefix, NULL);
 	update_string(&cfg.tab_label, NULL);
 	update_string(&cfg.tab_suffix, NULL);
@@ -146,12 +160,19 @@ conf_teardown(void)
 	update_string(&cfg.shell, NULL);
 	update_string(&cfg.shell_cmd_flag, NULL);
 
+	/* Non-zero history size has implications for views and status, so reset
+	 * it. */
+	cfg.history_len = 0;
+
 	cfg.dot_dirs = 0;
 
 	cfg.sizefmt.base = 0;
 	cfg.sizefmt.precision = 0;
 	cfg.sizefmt.ieci_prefixes = 0;
 	cfg.sizefmt.space = 0;
+
+	matcher_free(cfg.wild_inc);
+	cfg.wild_inc = NULL;
 }
 
 void
@@ -249,9 +270,13 @@ view_setup(view_t *view)
 	strcpy(view->curr_dir, "/path");
 	update_string(&view->custom.orig_dir, NULL);
 
-	view->sort[0] = SK_BY_NAME;
-	memset(&view->sort[1], SK_NONE, sizeof(view->sort) - 1);
-	memcpy(view->sort_g, view->sort, sizeof(view->sort_g));
+	view_set_sort(view->sort, SK_BY_NAME, SK_NONE);
+	view_set_sort(view->sort_g, SK_BY_NAME, SK_NONE);
+
+	/* The code assumes that this field is initialized.  At OS X and other
+	 * BSD-like refuse to compile empty regular expression. */
+	assert_success(regexp_compile(&view->primary_group, ".*", REG_ICASE));
+	view->primary_group_set = 1;
 
 	view->custom.entry_count = 0;
 	view->custom.entries = NULL;
@@ -270,6 +295,15 @@ view_teardown(view_t *view)
 	view->last_search[0] = '\0';
 }
 
+void
+view_set_sort(signed char sort_keys[], signed char primary_key,
+		signed char secondary_key)
+{
+	memset(sort_keys, SK_NONE, SK_COUNT);
+	sort_keys[0] = primary_key;
+	sort_keys[1] = secondary_key;
+}
+
 void
 columns_setup_column(int id)
 {
@@ -287,6 +321,7 @@ columns_teardown(void)
 {
 	columns_clear_column_descs();
 	columns_set_line_print_func(NULL);
+	columns_set_line_match_func(NULL);
 }
 
 void
@@ -658,6 +693,14 @@ check_compare_invariants(int expected_len)
 	}
 }
 
+void
+check_bg_jobs(void)
+{
+	/* Won't hurt to exercise code for displaying errors which depends on thread
+	 * synchronization. */
+	bg_check(/*show_errors=*/1);
+}
+
 void
 wait_for_bg(void)
 {
@@ -682,7 +725,7 @@ wait_for_all_bg(void)
 	setvar("v:jobcount", var);
 	var_free(var);
 
-	bg_check();
+	check_bg_jobs();
 	while(bg_jobs != NULL)
 	{
 		if(++counter > 100)
@@ -692,10 +735,39 @@ wait_for_all_bg(void)
 		}
 
 		usleep(5000);
-		bg_check();
+		check_bg_jobs();
 	}
 }
 
+int
+wait_for_job(struct bg_job_t *job)
+{
+	assert_non_null(job);
+
+	bg_job_incref(job);
+
+	int counter = 0;
+	while(bg_job_is_running(job))
+	{
+		usleep(5000);
+		check_bg_jobs();
+		if(++counter > 100)
+		{
+			assert_fail("Waiting for too long.");
+			return -1;
+		}
+	}
+
+	/* When the job is marked as not running, the callback might not yet been
+	 * dispatched, so call check_bg_jobs() once again to be sure. */
+	check_bg_jobs();
+
+	int exit_code = job->exit_code;
+	bg_job_decref(job);
+
+	return exit_code;
+}
+
 void
 file_is(const char path[], const char *lines[], int nlines)
 {
@@ -754,7 +826,7 @@ stub_colmgr(void)
 		.init_pair = &init_pair_stub,
 		.pair_content = &pair_content_stub,
 		.pair_in_use = &pair_in_use_stub,
-		.move_pair = &move_pair_stub,
+		.pair_moved = &pair_moved_stub,
 	};
 	colmgr_init(&colmgr_conf);
 }
@@ -780,7 +852,7 @@ pair_in_use_stub(int pair)
 }
 
 static void
-move_pair_stub(int from, int to)
+pair_moved_stub(int from, int to)
 {
 }
 
